use axum::{
    extract::multipart::MultipartError,
    http::{uri::InvalidUri, StatusCode},
    response::IntoResponse,
};
use log::error;
use thiserror::Error;
use tokio::task::JoinError;

#[derive(Debug, Error)]
pub enum Error {
    #[error("render error: {0:?}")]
    Render(#[from] sailfish::RenderError),

    #[error("invalid uri: {0:?}")]
    InvalidUri(#[from] InvalidUri),

    #[error("game not found")]
    NotFound,

    #[error("game broken")]
    GameBroken,

    #[error("game already started")]
    GameAlreadyStarted,

    #[error("field already submitted")]
    QuestionAlreadySubmitted,

    #[error("player not found")]
    PlayerNotFound,

    #[error("stream lagged")]
    StreamLagged(u64),

    #[error("multipart: {0:?}")]
    Multipart(#[from] MultipartError),

    #[error("quizfile not found")]
    QuizFileNotFound,

    #[error("toml")]
    Toml(#[from] toml::de::Error),

    #[error("forbidden")]
    Forbidden,

    #[error("invalid input")]
    InvalidInput,

    #[error("Prometheus: {0:?}")]
    Prometheus(std::fmt::Error),

    #[error("join error")]
    Join(#[from] JoinError),
}

impl IntoResponse for Error {
    fn into_response(self) -> axum::response::Response {
        match self {
            Self::Forbidden => (StatusCode::FORBIDDEN, "forbidden").into_response(),
            Self::Toml(_) => (StatusCode::OK, "invalid toml syntax").into_response(),
            Self::QuizFileNotFound => (StatusCode::OK, "quizfile not found").into_response(),
            Self::PlayerNotFound => (StatusCode::BAD_REQUEST, "player not found").into_response(),
            Self::QuestionAlreadySubmitted => {
                (StatusCode::BAD_REQUEST, "field already submitted").into_response()
            }
            Self::GameAlreadyStarted => {
                (StatusCode::BAD_REQUEST, "game already started").into_response()
            }
            Self::NotFound => (StatusCode::NOT_FOUND, "game not found").into_response(),
            _ => {
                error!("{:?}", self);
                (StatusCode::INTERNAL_SERVER_ERROR, "internal server error").into_response()
            }
        }
    }
}
