use std::{
    collections::{HashMap, HashSet},
    fmt::Debug,
    ops::Deref,
    sync::Arc,
    time::Duration,
};

use prometheus_client::{
    collector::Collector,
    encoding::{EncodeLabelSet, EncodeMetric},
    metrics::{
        family::Family,
        gauge::{ConstGauge, Gauge},
    },
};
use qrcode::{render::svg, QrCode};
use rand::{distributions, Rng};
use sailfish::TemplateOnce;
use serde::Deserialize;
use tokio::{
    select,
    sync::{broadcast, RwLock},
};

use crate::{
    error::Error, HandlerResult, PlayTemplate, PlayerState, Question, Quiz, ViewTemplate,
    ViewerState,
};

#[derive(Debug, Clone, PartialEq, Eq, Hash, Deserialize)]
pub struct GameId(Arc<str>);
#[derive(Debug, Clone, PartialEq, Eq, Hash, Deserialize)]
pub struct PlayerId(Arc<str>);

impl Deref for GameId {
    type Target = str;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl From<String> for GameId {
    fn from(value: String) -> Self {
        Self(value.into())
    }
}

impl GameId {
    pub fn random() -> Self {
        Self(
            rand::thread_rng()
                .sample_iter(distributions::Alphanumeric)
                .take(8)
                .map(char::from)
                .collect::<String>()
                .into(),
        )
    }
}

impl Deref for PlayerId {
    type Target = str;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl From<String> for PlayerId {
    fn from(value: String) -> Self {
        Self(value.into())
    }
}

impl PlayerId {
    pub fn random() -> Self {
        Self(
            rand::thread_rng()
                .sample_iter(distributions::Alphanumeric)
                .take(32)
                .map(char::from)
                .collect::<String>()
                .into(),
        )
    }
}

#[derive(Clone, Debug)]
pub enum GameState {
    NotStarted,
    Answering(u32),
    Result(u32),
    Completed,
}

pub struct Game {
    pub id: GameId,
    pub owner: String,
    pub state: Arc<RwLock<GameState>>,
    pub quiz: Quiz,
    pub players: HashSet<PlayerId>,
    pub on_state_update: broadcast::Sender<()>,
    pub on_submission: broadcast::Sender<()>,
    pub questions: Vec<Box<dyn Question>>,
}

impl Game {
    pub fn new(id: GameId, owner: String, quiz: Quiz) -> Self {
        Self {
            id,
            owner,
            questions: quiz
                .questions
                .iter()
                .map(|x| x.clone().into())
                .collect::<Vec<_>>(),
            quiz,
            state: Arc::new(RwLock::new(GameState::NotStarted)),
            players: HashSet::new(),
            on_state_update: broadcast::channel(16).0,
            on_submission: broadcast::channel(16).0,
        }
    }

    pub async fn next(&mut self) {
        let mut state = self.state.write().await;
        // game state machine
        *state = match *state {
            GameState::NotStarted => GameState::Answering(0),
            GameState::Answering(field) => GameState::Result(field),
            GameState::Result(field) if (field as usize) + 1 < self.quiz.questions.len() => {
                GameState::Answering(field + 1)
            }
            GameState::Result(_) => GameState::Completed,
            GameState::Completed => GameState::Completed,
        };

        self.on_state_update.send(());

        // automatically show results after x seconds, cancel when state is changed otherwise
        if let GameState::Answering(i) = *state {
            let state = self.state.clone();
            let state_update_tx = self.on_state_update.clone();
            let mut state_update_rx = self.on_state_update.subscribe();
            let wait_for = self.quiz.wait_for;
            tokio::spawn(async move {
                select! {
                    _ = tokio::time::sleep(Duration::from_secs(wait_for)) => {
                        let mut state = state.write().await;
                        *state = GameState::Result(i);
                        state_update_tx.send(());
                    },
                    _ = state_update_rx.recv() => {}
                }
            });
        }
    }

    pub async fn player_view(&self, player_id: &PlayerId, htmx: bool) -> HandlerResult<String> {
        if !self.players.contains(player_id) {
            return Err(Error::PlayerNotFound);
        }

        let state = self.state.read().await;

        let player_state = match *state {
            GameState::NotStarted => PlayerState::NotStarted,
            GameState::Answering(i)
                if !self.questions[i as usize].has_answered(player_id).await? =>
            {
                let inner_body = self.questions[i as usize]
                    .render_player(player_id, false)
                    .await?;
                PlayerState::Answering { inner_body }
            }
            GameState::Answering(i) => PlayerState::Waiting(i),
            GameState::Result(i) => {
                let inner_body = self.questions[i as usize]
                    .render_player(player_id, true)
                    .await?;
                PlayerState::Result { inner_body }
            }
            GameState::Completed => {
                let mut correct = 0;
                for question in self.questions.iter() {
                    if question.answered_correctly(player_id).await? {
                        correct += 1;
                    }
                }

                PlayerState::Completed(correct as f32 / (self.quiz.questions.len() as f32))
            }
        };

        Ok(PlayTemplate {
            htmx,
            id: &self.id,
            player_id,
            state: player_state,
        }
        .render_once()?)
    }

    pub async fn handle_answer(
        &mut self,
        player_id: &PlayerId,
        values: &HashMap<String, String>,
    ) -> HandlerResult<()> {
        if !self.players.contains(player_id) {
            return Err(Error::PlayerNotFound);
        }

        let state = self.state.read().await;

        if let GameState::Answering(i) = *state {
            self.questions[i as usize]
                .handle_answer(player_id.clone(), values)
                .await?;

            self.on_submission.send(());

            if self.questions[i as usize].answer_count().await? as usize == self.players.len() {
                drop(state);
                self.next().await;
            }
        }

        Ok(())
    }

    pub async fn viewer_view(&self, htmx: bool, base_url: &str) -> HandlerResult<String> {
        let state = self.state.read().await;

        let viewer_state = match *state {
            GameState::NotStarted => {
                let url = format!("{}/{}", base_url, &self.id.deref());
                let img = QrCode::new(&url).expect("");
                let img = img.render::<svg::Color>().build();
                ViewerState::NotStarted((self.players.len() as u32, img, url))
            }
            GameState::Answering(i) => {
                let inner_body = self.questions[i as usize].render_viewer(false).await?;

                ViewerState::Answering { inner_body }
            }
            GameState::Result(i) => {
                let inner_body = self.questions[i as usize].render_viewer(true).await?;

                ViewerState::Result {
                    last_question: (i + 1) as usize == self.quiz.questions.len(),
                    inner_body,
                }
            }

            GameState::Completed => ViewerState::Completed,
        };

        Ok(ViewTemplate {
            htmx,
            id: &self.id,
            state: viewer_state,
        }
        .render_once()?)
    }
}

#[derive(EncodeLabelSet, PartialEq, Eq, Hash, Clone, Debug)]
pub struct GameLabels {
    game: String,
}

pub struct RunningGamesCollector {
    games: Arc<RwLock<HashMap<GameId, Game>>>,
}

impl RunningGamesCollector {
    pub fn new(games: Arc<RwLock<HashMap<GameId, Game>>>) -> Self {
        Self { games }
    }
}

impl Debug for RunningGamesCollector {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("GameCollector").finish()
    }
}

impl Collector for RunningGamesCollector {
    fn encode(
        &self,
        mut encoder: prometheus_client::encoding::DescriptorEncoder,
    ) -> Result<(), std::fmt::Error> {
        let games = self.games.blocking_read();
        let running_games = ConstGauge::new(games.len() as i64);

        drop(games);

        let running_games_encoder = encoder.encode_descriptor(
            "ars_running_games",
            "number of running games",
            None,
            running_games.metric_type(),
        )?;
        running_games.encode(running_games_encoder)?;
        Ok(())
    }
}

pub struct GameParticipantsCollector {
    games: Arc<RwLock<HashMap<GameId, Game>>>,
}

impl GameParticipantsCollector {
    pub fn new(games: Arc<RwLock<HashMap<GameId, Game>>>) -> Self {
        Self { games }
    }
}

impl Debug for GameParticipantsCollector {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("GameCollector").finish()
    }
}

impl Collector for GameParticipantsCollector {
    fn encode(
        &self,
        mut encoder: prometheus_client::encoding::DescriptorEncoder,
    ) -> Result<(), std::fmt::Error> {
        let games = self.games.blocking_read();

        let participants = Family::<GameLabels, Gauge>::default();

        games.iter().for_each(|(id, game)| {
            participants
                .get_or_create(&GameLabels {
                    game: id.0.to_string(),
                })
                .set(game.players.len() as i64);
        });

        drop(games);

        let participants_encoder = encoder.encode_descriptor(
            "ars_game_participants",
            "number of participants for a game",
            None,
            participants.metric_type(),
        )?;
        participants.encode(participants_encoder)?;

        Ok(())
    }
}
