use std::collections::HashMap;

use axum::async_trait;
use sailfish::TemplateOnce;

use crate::{error::Error, game::PlayerId, Question, SingleChoice};

pub struct SingleChoiceQuestion {
    inner: SingleChoice,
    submissions: HashMap<PlayerId, u32>,
}

impl SingleChoiceQuestion {
    pub fn new(single_choice: SingleChoice) -> Self {
        Self {
            inner: single_choice,
            submissions: HashMap::new(),
        }
    }
}

#[async_trait]
impl Question for SingleChoiceQuestion {
    async fn render_player(
        &self,
        player_id: &PlayerId,
        show_result: bool,
    ) -> Result<String, Error> {
        if show_result {
            let player_sub_index = self.submissions.get(player_id);

            let player_sub_value = player_sub_index.map(|x| &*self.inner.answers[*x as usize]);

            Ok(ResultTemplate {
                is_correct: Some(&self.inner.correct) == player_sub_index,
                correct_answer: &self.inner.answers[self.inner.correct as usize],
                player_answer: player_sub_value,
            }
            .render_once()?)
        } else {
            Ok(PlayerTemplate {
                name: &self.inner.name,
                player_id,
                answers: &self.inner.answers,
            }
            .render_once()?)
        }
    }

    async fn handle_answer(
        &mut self,
        player_id: PlayerId,
        values: &HashMap<String, String>,
    ) -> Result<(), Error> {
        let value = values
            .get("value")
            .ok_or(Error::InvalidInput)?
            .parse::<u32>()
            .map_err(|_| Error::InvalidInput)?;

        if value as usize >= self.inner.answers.len() {
            return Err(Error::InvalidInput);
        }

        if self.submissions.contains_key(&player_id) {
            return Err(Error::QuestionAlreadySubmitted);
        }

        self.submissions.insert(player_id, value);

        Ok(())
    }

    async fn has_answered(&self, player_id: &PlayerId) -> Result<bool, Error> {
        Ok(self.submissions.get(player_id).is_some())
    }

    async fn answered_correctly(&self, player_id: &PlayerId) -> Result<bool, Error> {
        Ok(self
            .submissions
            .get(player_id)
            .map(|x| *x == self.inner.correct)
            .unwrap_or_default())
    }

    async fn answer_count(&self) -> Result<u32, Error> {
        Ok(self.submissions.len() as u32)
    }

    async fn render_viewer(&self, show_result: bool) -> Result<String, Error> {
        Ok(ViewerTemplate {
            show_result,
            name: &self.inner.name,
            total_submissions: self.submissions.len() as u32,
            submissions: &self.submissions.iter().fold(
                vec![0; self.inner.answers.len()].into_boxed_slice(),
                |mut acc, (_, v)| {
                    acc[*v as usize] += 1;
                    acc
                },
            ),
            submissions_correct: &self
                .submissions
                .iter()
                .filter(|(_, v)| **v == self.inner.correct)
                .fold(
                    vec![0; self.inner.answers.len()].into_boxed_slice(),
                    |mut acc, (_, v)| {
                        acc[*v as usize] += 1;
                        acc
                    },
                ),
            submissions_wrong: &self
                .submissions
                .iter()
                .filter(|(_, v)| **v != self.inner.correct)
                .fold(
                    vec![0; self.inner.answers.len()].into_boxed_slice(),
                    |mut acc, (_, v)| {
                        acc[*v as usize] += 1;
                        acc
                    },
                ),
            correct_answer: &self.inner.answers[self.inner.correct as usize],
            answers: &self.inner.answers,
        }
        .render_once()?)
    }
}

#[derive(TemplateOnce)]
#[template(path = "single_choice/player.stpl")]
struct PlayerTemplate<'a> {
    name: &'a str,
    player_id: &'a str,
    answers: &'a [Box<str>],
}

#[derive(TemplateOnce)]
#[template(path = "single_choice/result.stpl")]
struct ResultTemplate<'a> {
    is_correct: bool,
    correct_answer: &'a str,
    player_answer: Option<&'a str>,
}

#[derive(TemplateOnce)]
#[template(path = "single_choice/viewer.stpl")]
struct ViewerTemplate<'a> {
    show_result: bool,
    name: &'a str,
    total_submissions: u32,
    submissions: &'a [u32],
    submissions_correct: &'a [u32],
    submissions_wrong: &'a [u32],
    correct_answer: &'a str,
    answers: &'a [Box<str>],
}
